# Copyright 2022 Google, LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

package Google::Auth::ComputeEngine;

use strict;
use warnings;

use Moo;

our $VERSION = '0.02';

sub on_gce {
    my ( $class, %options ) = @_;
    # Return false for now until metadata server detection is implemented
    return 0;
}

1;

